#include <stdio.h>
#include <gmp.h>
#include <mpi.h>
#include <math.h>
#include <limits.h>
#define MAXmaxp 1024

int n,maxp,over=-1,primes[MAXmaxp],extraprimes[2*MAXmaxp],pid=0;

int ref(int p) 
{
	int i;
	for(i=0;i<maxp;i++)
		if(p==primes[i])
			break;
	return i;
}

long mod(long x, long y)
{
        if(x>0)
                return x%y;
        while(x<0)
                x+=y;
        return x;
}

int closestprime(double n,FILE *fp)
{
	int cprime;
	fscanf(fp,"%d",&cprime);
	while(n-log(cprime)>0)
		fscanf(fp,"%d",&cprime);
	return cprime;
}

int printmat(int **A)
{
	int i,j;
	for(i=0;i<n;i++) {
		printf("%d:\t",pid);
		for(j=0;j<n+1;j++)
			printf("%d\t",A[i][j]);
		printf("\n");
	}
}

int euclid(int a, int b)
{
	int x,v,y,u,tx,ty,mod=b;
	x=v=1;
	y=u=0;
	while(b!=0) {
		int q=a/b;
		int r=a%b;
		a=b;b=r;
		tx=x;ty=y;
		x=u;y=v;
		u=tx-q*u;v=ty-q*v;
	}
	return (x<0)?x+mod:x; 
}

int fail(int p)
{
	fflush(stdout);
	MPI_Send(&pid,1,MPI_INT,0,1,MPI_COMM_WORLD);
	childproc();
	return -1;
}

int intgauss(int **A,int *b,int *x, int *d, int p)
{
        int **B;
        int i,ip,j,q,k,t,s,S;

	B=(int **)calloc(n,sizeof(int *));
	B[0]=(int *)malloc(sizeof(int)*n*(n+1));
	for(i=1;i<n;i++)
		B[i]=B[0]+i*(n+1);

	S=1;
        for(j=0;j<n;j++)
                for(k=0;k<n+1;k++)
                        if(k==n)
                                B[j][k]=b[j];
                        else
                                B[j][k]=A[j][k];
	//printmat(B);
        for(i=1;i<=n;i++) {
                int max=i,q1,R;
                for(k=j+1;k<=n;k++)
                        t+=mod(B[j-1][k-1]*x[k-1],p);
                for(j=i+1;j<=n;j++)
                        if(abs(B[j-1][i-1])>abs(B[max-1][i-1]))
                                max=j;
                if(i!=max)
                        S=-S;
                for(k=i;k<=n+1;k++) {
                        R=B[i-1][k-1];
                        B[i-1][k-1]=B[max-1][k-1];
                        B[max-1][k-1]=R;
                }
		q=euclid(B[i-1][i-1],p);
                if((q*B[i-1][i-1])%p!=1)
                        return fail(p);
                for(j=i+1;j<=n;j++)
                        for(k=n+1;k>=i;k--) 
                                B[j-1][k-1]=mod(B[j-1][k-1]-mod(B[i-1][k-1]*B[j-1][i-1],p)*q,p);
        }
        for(j=n;j>=1;j--) {
		int t=0;
                for(k=j+1;k<=n;k++)
                        t+=mod(B[j-1][k-1]*x[k-1],p);
		q=euclid(B[j-1][j-1],p);
                if((q*B[j-1][j-1])%p!=1) {
                        //printf("%d: HERE fail 1\n", pid);
                        return fail(p);
                }
                x[j-1]=mod((B[j-1][n]-t)*q,p);
        }
        for(i=1;i<=n;i++) {
		S=S*B[i-1][i-1];
		S=mod(S,p);
	}
	*d=S;
	for(i=0;i<n;i++)
		x[i]=mod(x[i]**d,p);
	return 0;
}

int newman(mpz_t **A,mpz_t *b)
{
	mpz_t temp1,temp2,m;
	int processes=0,i,ip,j,primeprime[MAXmaxp],q,anid,tempd[MAXmaxp];
	int *tempb,**tempA,**tempy,singcount=0;
	char a[700];

	mpz_init(temp1);
	mpz_init(temp2);
	mpz_init(m);
	mpz_set_si(m,1);

	tempb=(int *)calloc(n,sizeof(int));
	tempA=(int **)calloc(n,sizeof(int *));
	tempy=(int **)calloc(maxp,sizeof(int *));	
	for(i=0;i<n;i++)
		tempA[i]=(int *)calloc(n,sizeof(int));
	for(i=0;i<maxp;i++)
		tempy[i]=(int *)calloc(n,sizeof(int));

	//printf("newman entered...\n");
	for(ip=0,anid=1;ip<maxp;ip++,anid++) {
		for(j=0;j<n;j++) {
			mpz_mod_ui(temp1,b[j],primes[ip]);
			tempb[j]=mpz_get_si(temp1);
			for(q=0;q<n;q++) {
				mpz_mod_ui(temp1,A[j][q],primes[ip]);
				tempA[j][q]=mpz_get_si(temp1);
			}
		}
		for(i=0;i<n;i++) {
			MPI_Send(tempA[i],n,MPI_INT,anid,0,MPI_COMM_WORLD);
                }
		MPI_Send(tempb,n,MPI_INT,anid,0,MPI_COMM_WORLD);
		MPI_Send(&(primes[ip]),1,MPI_INT,anid,0,MPI_COMM_WORLD);
		processes++;
	}
	while(processes>0) {
		int flag;
		MPI_Status status;
		//printf("processes: %d\n",processes);
		//check for new returns
		MPI_Probe(MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&status);
		//printf("probe unblocked\n");
		if(status.MPI_TAG==0) {
			int p,r, src;
			//printf("receiving result...\n");
                        src=status.MPI_SOURCE;
			MPI_Recv(&p,1,MPI_INT,src,0,MPI_COMM_WORLD,&status);
			r=ref(p);
			MPI_Recv(tempy[r],n,MPI_INT,src,0,MPI_COMM_WORLD,&status);
			MPI_Recv(&(tempd[r]),1,MPI_INT,src,0,MPI_COMM_WORLD,&status);
			mpz_mul_ui(m,m,p);
			processes--;
		}
		//check for singular stuff
		if(status.MPI_TAG==1) {
			int id,src;
			singcount++;
                        src=status.MPI_SOURCE;
			MPI_Recv(&id,1,MPI_INT,src,1,MPI_COMM_WORLD,&status);
			printf("recieved singular...%d %d\n",primes[id-1],extraprimes[over+1]);
			over++;
			primes[id-1]=extraprimes[over];
			for(j=0;j<n;j++) {
				mpz_mod_ui(temp1,b[j],primes[id-1]);
				tempb[j]=mpz_get_si(temp1);
				for(q=0;q<n;q++) {
					mpz_mod_ui(temp1,A[j][q],primes[id-1]);
					tempA[j][q]=mpz_get_si(temp1);
				}
			}
			for(i=0;i<n;i++)
				MPI_Send(tempA[i],n,MPI_INT,id,0,MPI_COMM_WORLD);
			MPI_Send(tempb,n,MPI_INT,id,0,MPI_COMM_WORLD);
			MPI_Send(&(primes[id-1]),1,MPI_INT,id,0,MPI_COMM_WORLD);
		}
	}
	fprintf(stderr,"singular: %d\n",singcount);
	fflush(stderr);
	//printf("all data collected\n");
	for(i=0;i<maxp;i++) {
		for(primeprime[i]=1;primeprime[i]<primes[i];primeprime[i]++) {
			mpz_set(temp1,m);
			mpz_tdiv_q_ui(temp1,temp1,primes[i]);
			mpz_mul_ui(temp1,temp1,primeprime[i]);
			mpz_mod_ui(temp1,temp1,primes[i]);
			if(mpz_get_si(temp1)==1) {
				//printf("%d: %d\n",primes[i],primeprime[i]);
				break;
			}
		}
		if(primeprime[i]==primes[i])
			printf("replace %d -- it doesn't invert\n",primes[i]);
	}
	//printf("primeprime computed\n");
	mpz_set_si(temp1,0);
	for(i=0;i<maxp;i++) {
		mpz_set_si(temp2,tempd[i]);
		mpz_mul(temp2,temp2,m);
		mpz_tdiv_q_ui(temp2,temp2,primes[i]);
		mpz_mul_ui(temp2,temp2,primeprime[i]);
		mpz_add(temp1,temp1,temp2);
	}
	mpz_tdiv_q_ui(temp2,m,2);
	mpz_mod(temp1,temp1,m);
	if(mpz_cmp(temp1,temp2)>0)
		mpz_sub(temp1,temp1,m);
	printf("d: %s\n",mpz_get_str(a,10,temp1));
	for(j=0;j<n;j++) {
		mpz_set_si(temp1,0);
		mpz_set_si(temp2,0);
		for(i=0;i<maxp;i++) {
			mpz_set_si(temp2,tempy[i][j]);
			mpz_mul(temp2,temp2,m);
			mpz_tdiv_q_ui(temp2,temp2,primes[i]);
			mpz_mul_ui(temp2,temp2,primeprime[i]);
			mpz_add(temp1,temp1,temp2);
		}
		mpz_tdiv_q_ui(temp2,m,2);
		mpz_mod(temp1,temp1,m);
		if(mpz_cmp(temp1,temp2)>0)
			mpz_sub(temp1,temp1,m);
		printf("y[%d]: %s\n",j,mpz_get_str(a,10,temp1));
		fflush(stdout);
	}
}

int childproc()
{
		MPI_Status status;
		int **A,*b,*x;
                int i,j,d,p;
		A=(int **)calloc(n,sizeof(int *));
		b=(int *)calloc(n,sizeof(int));
		x=(int *)calloc(n,sizeof(int));
		for(i=0;i<n;i++)
			A[i]=(int *)calloc(n,sizeof(int));
                for(i=0;i<n;i++)
                        MPI_Recv(A[i],n,MPI_INT,0,0,MPI_COMM_WORLD,&status);
                MPI_Recv(b,n,MPI_INT,0,0,MPI_COMM_WORLD,&status);
                MPI_Recv(&p,1,MPI_INT,0,0,MPI_COMM_WORLD,&status);
		//printf("%d recieved data\n",pid);
                if(!intgauss(A,b,x,&d,p)) {
		//printf("%d got answer\n",pid);
	                MPI_Send(&p,1,MPI_INT,0,0,MPI_COMM_WORLD);
       	         	MPI_Send(x,n,MPI_INT,0,0,MPI_COMM_WORLD);
                	MPI_Send(&d,1,MPI_INT,0,0,MPI_COMM_WORLD);
		}
		//printf("%d child exiting\n",pid);
}

int main(int argc, char **argv)
{
	MPI_Status status;
	mpz_t **A,*b;
	int i,j,ierr;
	char str[100];
        int nprocs;
        FILE *infile,*fp;
        char infile_name[1000];

	ierr=MPI_Init(&argc,&argv);
	ierr=MPI_Comm_rank(MPI_COMM_WORLD,&pid);

	MPI_Comm_size(MPI_COMM_WORLD,&maxp);
	maxp--;

	if(pid==0) {
		if (argc != 2) {
			printf("Command-line error, input file name needed.\n");
			MPI_Abort(MPI_COMM_WORLD,2);
		}
		if ((infile = fopen(argv[1],"r")) == NULL) {
			printf("Could not open input file %s.\n", argv[1]);
			MPI_Abort(MPI_COMM_WORLD,2);
		}
		fscanf(infile,"%d",&n);
	}
	MPI_Bcast(&n,1,MPI_INT,0,MPI_COMM_WORLD);
	if(pid==0) {
		mpz_t maxa,maxb,temp;
		double u,v,ntot;
		int q;

		A=(mpz_t **)calloc(n,sizeof(mpz_t *));
		b=(mpz_t *)calloc(n,sizeof(mpz_t));
		mpz_init(maxa);
		mpz_init(maxb);
		mpz_init(temp);

		mpz_set_ui(maxa,0);
		mpz_set_ui(maxb,0);

		for(i=0;i<n;i++)
			A[i]=(mpz_t *)calloc(n,sizeof(mpz_t));
		for(i=0;i<n;i++)
			for(j=0;j<n;j++) {
				mpz_init(A[i][j]);
				fscanf(infile,"%s",str);
				mpz_set_str(A[i][j],str,10);
				mpz_abs(temp,A[i][j]);
				if(mpz_cmp(temp,maxa)>0)
					mpz_set(maxa,temp);
			}
		for(i=0;i<n;i++) {
			mpz_init(b[i]);
			fscanf(infile, "%s",str);
			mpz_set_str(b[i],str,10);
			mpz_abs(temp,b[i]);
			if(mpz_cmp(temp,maxb)>0)
				mpz_set(maxb,temp);
		}
		//printf("data gathered, calling newman()...\n");
		fclose(infile);
		u=log(n)+(n-1)/2.0*log(n-1)+(n-1)*log(mpz_get_ui(maxa))+log(mpz_get_ui(maxb));
		v=n/2.0*log(n)+n*log(mpz_get_ui(maxa));
		if(u>v)
			ntot=log(2)+u;
		else
			ntot=log(2)+v;
		printf("necessary digits: %d %f\n",(int)floor(ntot)+1,ntot/maxp);
		fp=fopen("prime.data","r");
		printf("chosen primes:\t");
		primes[0]=closestprime(ntot/maxp,fp);
		printf("%d ",primes[0]);
		for(i=1;i<maxp;i++) {
			fscanf(fp,"%d",&(primes[i]));
			printf("%d ",primes[i]);
		}
		if(primes[i-1]>INT_MAX) {
			perror("exceeding word size\n");
			exit(-1);
		}
		for(i=maxp;i<3*maxp;i++,q++)
			fscanf(fp,"%d",&(extraprimes[i-maxp]));
		fclose(fp);
		printf("\n");
		newman(A,b);
	} else {
		//printf("task %d starting...\n",pid);
		childproc();
	}
	fflush(stdout);
	MPI_Barrier(MPI_COMM_WORLD);
	ierr=MPI_Finalize();
	return 0;
}
